from scenario_gym.entity.base import Entity
from scenario_gym.entity.batch import BatchReplayEntity
from scenario_gym.entity.misc import MiscObject, MiscObjectCatalogEntry
from scenario_gym.entity.pedestrian import Pedestrian, PedestrianCatalogEntry
from scenario_gym.entity.vehicle import Vehicle, VehicleCatalogEntry

DEFAULT_ENTITY_TYPES = (Vehicle, Pedestrian, MiscObject)
